// *****************************************************************************
//  void harderr(int (*fptr)())   - establishes a hardware error handler for    
//                                  int 24, utilizing a DPMI real-mode callback.
//                                  Both PM and a RM vectors are set up since   
//                                  DPMI is not clear as to which one will get  
//                                  called.  RTM appears to pass all int 24's to
//                                  RM.                                         
//  void end_harderr(void)        - removes harware interrupt handler.  If not  
//                                  not called explicitly, it will be called as 
//                                  an exit function. (DOS also restores the RM 
//                                  RM vector, but DPMI probably does not       
//                                  restore the PM vector)                      
//  Description:   harderr establishes a hardware error handler for the current 
//                 program. This handler is invoked whenever an interrupt 0x24  
//                 occurs. The function pointed to by fptr will be called when  
//                 such an interrupt occurs. The handler function will be called
//                 with the following arguments:                                
//                                                                              
//                 int handler(int errval, int ax, int bp, int si);             
//                  - errval is the error code set in the DI register by MS-DOS.
//                  - ax, bp, and si are the values MS-DOS sets for the AX, BP, 
//                    and SI registers, respectively.                           
//                    - ax indicates whether a disk error or other device error 
//                      was encountered. If ax is non-negative, a disk error was
//                      encountered; otherwise, the error was a device error.   
//                      For a disk error, ax ANDed with 0x00FF will give the    
//                      failing drive number (1 = A, 2 = B, etc.).              
//                    - bp and si together point to the device driver header of 
//                      the failing driver. (SEG:offset, NOT sel:offset!)       
//                  - harderr establishes a DOS interrupt handler that calls the
//                    function, utilizing a DPMI real-mode callback.            
//                  - Dos calls are NOT safe to call during execution of the    
//                    error handler.                                            
//                  - The error handler must return with a value of 1 for retry,
//                    0 for ignore, and 3 for fail. The handler cannot force an 
//                    abort.                                                    
// NOTES: - Only works for dpmi16 - If you modify it for dpmi32, please post it.
//          I don't use dpmi32, but I have noticed posts saying that dpmi32     
//          automatically fails all int24's instead of passing them through...  
//        - Requires Pat Reilly's pdr1 library, available in the compuserve dpmi
//          library - If you don't have it, its a set of C wrappers and structs 
//          for the dpmi calls. (The ones the powerpack manual says you         
//          shouldn't need)                                                     
//        - Requires tasm - It's implemented in C only because I originally     
//          thought I could do most of it in C. Turns out its mostly assembly,  
//          which I'm pretty rusty at...                                        
//                                                                              
// AUTHOR:  Eric Noel                                                           
// RELEASE: 5/28/95                                                             
//                                                                              
// This software is hereby released into the public domain - it is "freeware".  
// You may use it for any purpose except that it may not be sold unless it is   
// incorporated into your application that is distributed as an executable.     
//                                                                              
//                                 DISCLAIMER:                                  
// -----------------------------------------------------------------------------
//      THIS SOFTWARE IS DISTRIBUTED "AS IS" AND WITHOUT WARRANTIES AS TO       
//  PERFORMANCE OF MERCHANTABILITY OR ANY OTHER WARRANTIES WHETHER EXPRESSED OR 
//   IMPLIED.  BECAUSE OF THE VARIOUS HARDWARE AND SOFTWARE ENVIRONMENTS INTO   
//    WHICH THIS PRODUCT MAY BE PUT, NO WARRANTY OF FITNESS FOR A PARTICULAR    
//    PURPOSE IS OFFERED.  THE USER MUST ASSUME THE ENTIRE RISK OF USING THE    
//  SOFTWARE.  I WILL NOT BE RESPONSIBLE FOR ANY DAMAGE OR LOSS OF TIME, DATA,  
//         HARDWARE, OR SOFTWARE YOU INCUR THROUGH USE OF THIS PRODUCT.         
// -----------------------------------------------------------------------------
//                                                                              
// This software was designed for my own use.  As a result of several requests  
// for this type of a function, I have cleaned it up and added comments so that 
// you can use it as-is, or as a callback example.  However, I cannot provide   
// personal design or usage support.  If you have questions or problems with it 
// or DPMI in general, post a message in section 7 (DPMI) of the BCPPDOS forum. 
// If I don't/cannot answer your question, someone else probably will. If you   
// find any bugs or have a better solution, please post your version to the dpmi
// library on compuserve.                                                       
//                                                                              
// As an aside, there is absolutely NO excuse for Borland's decision not to     
// implement this function in the rtl. Over the last month or so I've seen quite
// a few threads (on compuserve) from programmers wondering how to go about     
// porting their int24 functions. Almost all professional dos programs use some 
// sort of GUI that requires int24 to be trapped.  Borland even managed to      
// include a handler in TV - Why would they believe that other GUI's would not  
// require it?  Doesn't make any sense to me.                                   
// *****************************************************************************
#pragma inline

#include <dos.h>
#include <mem.h>
#include "\mylib\pdr1.h"

// turn off dead store optimization
#pragma option -O-b 

#define DI_OFFSET 0x00
#define SI_OFFSET 0x04
#define BP_OFFSET 0x08
#define AX_OFFSET 0x1C
#define FL_OFFSET 0x20
#define IP_OFFSET 0x2A
#define CS_OFFSET 0x2C
#define SP_OFFSET 0x2E

PMRegs regbuf;

static  int (*Hfunc)(unsigned di, unsigned ax, unsigned bp, unsigned si);

// PM int 24 handler to call user error handler (Same as old method):           
#pragma argsused
static void interrupt far hentry(bp, di, si, ds, es, dx, cx, bx, ax)
{   ax = (*Hfunc)(di & 0xff, ax, bp, si);
}


// Real mode version, called as a dpmi callback:                                
//                                                                              
// From the dpmi specs:                                                         
//                                                                              
// ENTRY:   CS:IP -> This routine                                               
//          DS:SI -> Real mode SS:SP                                            
//          ES:DI -> Real mode call structure                                   
//          SS:SP -> Locked protected mode stack from dpmi host                 
//          Interrupts disabled                                                 
// EXIT:    ES:DI -> Real mode call stucture                                    
static void interrupt far hentry_from_real(void)
{  
asm {
   // The interrupt keyword forces all registers to be saved, and then the      
   // compiler throws in code that screws up bp and ss.  Restore registers to   
   // what they were upon entry:                                                
	pop	bp
	pop	edi
	pop	esi
	pop	ds
	pop	es
	pop	edx
	pop	ecx
	pop	ebx
	pop	eax
   // NOTE: The above works on BC v4.02 - If using a different version, compile 
   // this to asm and make sure the above 'pops' exactly match the 'push' codes 
   // generated by the compiler.                                                

   // Get and store return address for return to real mode, from real mode's    
   // stack (addressable through ds:si, and in seg:off form in regbuf).
	cld                      
   lodsw
   mov   es:[di+IP_OFFSET],ax    ;// Get the real modes IP
   lodsw
   mov   es:[di+CS_OFFSET],ax    ;// Get the real modes CS

   lodsw
   mov   es:[di+FL_OFFSET],ax    ;// Get the real modes FLAGS

   // This function was called initially as an interrupt, but it will be        
   // returned directly rather than by an iret, so the real mode's stack needs  
   // the interrupt flags, ip & cs data removed:                                
   add   es:[di+SP_OFFSET], 6    

   // Enable interrupts.  This code is NOT re-entrant, however, if the user     
   // function uses any dos calls that could generate another int24, the system 
   // will crash anyway:                                                        
   sti

   // Have to restore ds to Hfunc's, so that it can be accessed:                
   mov   ax,DGROUP
   mov   ds,ax
  
   // Save es:di in case the user handler trashes them:                         
   push  es
   push  di

   // Call the user handler:                                                    
   mov   ax,es:[di+SI_OFFSET]
   push  ax
   mov   ax,es:[di+BP_OFFSET]
   push  ax
   mov   ax,es:[di+AX_OFFSET]
   push  ax
   mov   ax,es:[di+DI_OFFSET]
	and	ax,255
   push  ax
	call	dword ptr Hfunc
	add	sp,8

   // Restore es:di                                                             
   pop   di
   pop   es

   // Set the ax value to be returned to dos:                                   
	mov	es:[di+AX_OFFSET],ax

   // Have to force an iret now or the compiler will try to pop the data it     
   // thinks is still on the stack:                                             
   leave
   iret
   }
}



RMAddress callback=0;
RMAddress RM_oldvect;
PMFarAddress PM_oldvect;

void end_harderr(void)
{  if (callback!=0)
   {  dpmi_setRMVect(0x24,RM_oldvect);
      dpmi_setPMVect(0x24,PM_oldvect);
      dpmi_freeCallBack(callback);
      callback=0;
      }
}
#pragma exit end_harderr

void  harderr(int (*fptr)())      
{  Hfunc = fptr;

   end_harderr();
   callback=dpmi_allocateCallBack(hentry_from_real,&regbuf);
   RM_oldvect=dpmi_getRMVect(0x24);
   dpmi_setRMVect(0x24,callback);

   PM_oldvect=dpmi_getPMVect(0x24);
   dpmi_setPMVect(0x24,hentry);
}



